from __future__ import annotations

import re
from dataclasses import dataclass
from typing import NewType, Protocol, cast

import pytest

from waku import WakuApplication, WakuFactory
from waku.di import Scope, contextual, scoped
from waku.ext.validation import ValidationError, ValidationExtension, ValidationRule
from waku.ext.validation.rules import DependenciesAccessible
from waku.modules import ModuleType, module


@dataclass
class A:
    pass


@dataclass
class B:
    a: A


C = NewType('C', A)


@dataclass
class D:
    c: C


def _impl() -> int:  # pragma: no cover
    return 1


@pytest.fixture
def rule() -> ValidationRule:
    return DependenciesAccessible()


class ApplicationFactoryFunc(Protocol):
    def __call__(
        self,
        root_module: ModuleType,
        *,
        strict: bool = True,
        context: dict[type, object] | None = None,
    ) -> WakuApplication: ...


@pytest.fixture
def application_factory(rule: ValidationRule) -> ApplicationFactoryFunc:
    def factory(
        root_module: ModuleType,
        *,
        strict: bool = True,
        context: dict[type, object] | None = None,
    ) -> WakuApplication:
        return WakuFactory(
            root_module,
            context=context or {},
            extensions=[ValidationExtension([rule], strict=strict)],
        ).create()

    return factory


@pytest.mark.parametrize(
    ('imports', 'exports', 'should_fail'),
    [
        (False, False, True),
        (False, True, True),
        (True, False, True),
        (True, True, False),
    ],
)
async def test_accessibility_import_export_matrix(
    imports: bool,
    exports: bool,
    should_fail: bool,
    application_factory: ApplicationFactoryFunc,
) -> None:
    """Test all combinations of import/export for dependency accessibility."""

    @module(providers=[scoped(A)], exports=[A] if exports else [])
    class AModule:
        pass

    @module(providers=[scoped(B)], imports=[AModule] if imports else [])
    class BModule:
        pass

    @module(imports=[AModule, BModule])
    class AppModule:
        pass

    application = application_factory(AppModule)
    if should_fail:
        with pytest.raises(ExceptionGroup) as exc_info:
            await application.initialize()
        b_module = application.registry.get(BModule)
        error_message = f'"{B!r}" from "{b_module!r}" depends on "{A!r}" but it\'s not accessible to it'
        error = cast(ValidationError, exc_info.value.exceptions[0].exceptions[0])
        assert str(error).startswith(error_message)
    else:
        await application.initialize()


async def test_accessible_with_exported_and_imported(application_factory: ApplicationFactoryFunc) -> None:
    """Test that dependency is accessible when exported and imported."""

    @module(providers=[scoped(A), scoped(_impl, provided_type=C)], exports=[A, C])
    class AModule:
        pass

    @module(providers=[scoped(B), scoped(D)], imports=[AModule])
    class BModule:
        pass

    @module(imports=[AModule, BModule])
    class AppModule:
        pass

    await application_factory(AppModule).initialize()


async def test_accessible_with_global_provider(application_factory: ApplicationFactoryFunc) -> None:
    """Test that global providers are accessible everywhere."""

    @module(providers=[scoped(A)], is_global=True)
    class AModule:
        pass

    @module(providers=[scoped(B)], imports=[AModule])
    class BModule:
        pass

    @module(imports=[AModule, BModule])
    class AppModule:
        pass

    await application_factory(AppModule).initialize()


@pytest.mark.parametrize('scope', [Scope.APP, Scope.REQUEST])
async def test_accessible_with_contextual_provider(
    rule: ValidationRule,
    scope: Scope,
) -> None:
    """Test that contextual providers are accessible if present in context."""

    @module(
        providers=[
            contextual(A, scope=scope),
            scoped(B),
        ],
        exports=[B],
    )
    class Module:
        pass

    @module(imports=[Module])
    class AppModule:
        pass

    application = WakuFactory(
        AppModule,
        context={A: A()},
        extensions=[ValidationExtension([rule])],
    ).create()
    await application.initialize()


async def test_accessible_with_application_providers(application_factory: ApplicationFactoryFunc) -> None:
    """Test that providers in the application module are accessible."""

    @module(providers=[scoped(B)], exports=[B])
    class BModule:
        pass

    @module(providers=[scoped(A)], imports=[BModule])
    class AppModule:
        pass

    application: WakuApplication = application_factory(AppModule)
    await application.initialize()


async def test_intra_module_access(application_factory: ApplicationFactoryFunc) -> None:
    """Test that providers can access each other within the same module without export."""

    @module(providers=[scoped(A), scoped(B)])
    class Module:
        pass

    @module(imports=[Module])
    class AppModule:
        pass

    await application_factory(AppModule).initialize()


@dataclass
class X:
    pass


@dataclass
class Y:
    pass


@dataclass
class Z:
    x: X
    y: Y


async def test_multiple_missing_dependencies(application_factory: ApplicationFactoryFunc) -> None:
    """Test that multiple missing dependencies are all reported."""

    @module(providers=[scoped(X), scoped(Y)])
    class XYModule:
        pass

    @module(providers=[scoped(Z)])
    class ZModule:
        pass

    @module(imports=[XYModule, ZModule])
    class AppModule:
        pass

    with pytest.raises(ExceptionGroup) as exc_info:
        await application_factory(AppModule).initialize()

    validation_errors = exc_info.value.exceptions[0].exceptions
    error_msgs = [str(e) for e in validation_errors]
    assert any('depends on' in msg and 'X' in msg for msg in error_msgs)
    assert any('depends on' in msg and 'Y' in msg for msg in error_msgs)


async def test_warning_mode(application_factory: ApplicationFactoryFunc) -> None:
    """Test that non-strict mode emits warnings instead of raising."""

    @module(providers=[scoped(A)])
    class AModule:
        pass

    @module(providers=[scoped(B)], imports=[])  # No import, should warn
    class BModule:
        pass

    @module(imports=[AModule, BModule])
    class AppModule:
        pass

    application = application_factory(AppModule, strict=False)

    b_module = application.registry.get(BModule)
    error_message = f'"{B!r}" from "{b_module!r}" depends on "{A!r}" but it\'s not accessible to it'
    with pytest.warns(Warning, match=re.escape(error_message)):
        await application.initialize()



from __future__ import annotations

from collections import OrderedDict, defaultdict
from typing import TYPE_CHECKING, Final

from waku.extensions import ModuleExtension, OnModuleConfigure
from waku.modules import Module, ModuleCompiler, ModuleMetadata, ModuleRegistry, ModuleType

if TYPE_CHECKING:
    from uuid import UUID

    from dishka.provider import BaseProvider

    from waku import DynamicModule


__all__ = ['ModuleRegistryBuilder']


class ModuleRegistryBuilder:
    def __init__(self, root_module_type: ModuleType, compiler: ModuleCompiler | None = None) -> None:
        self._compiler: Final = compiler or ModuleCompiler()
        self._root_module_type: ModuleType = root_module_type
        self._modules: dict[UUID, Module] = {}
        self._providers: list[BaseProvider] = []
        self._extensions: list[ModuleExtension] = []
        self._adjacency: dict[UUID, OrderedDict[UUID, str]] = defaultdict(OrderedDict)

    def build(self) -> ModuleRegistry:
        modules = self._collect_modules()
        root_module = self._register_modules(modules)
        return self._build_registry(root_module)

    def _collect_modules(self) -> list[tuple[ModuleType, ModuleMetadata]]:
        """Collect modules in post-order DFS."""
        stack: list[tuple[ModuleType | DynamicModule, bool]] = [(self._root_module_type, False)]
        post_order: list[tuple[ModuleType, ModuleMetadata]] = []
        visited: set[UUID] = set()
        while stack:
            current_type, processed = stack.pop()
            type_, metadata = self._compiler.extract_metadata(current_type)

            module_id = metadata.id
            if module_id in visited:
                continue

            if processed:
                post_order.append((type_, metadata))
                visited.add(module_id)
                continue

            stack.append((current_type, True))
            for imported in reversed(metadata.imports):
                _, imported_metadata = self._compiler.extract_metadata(imported)
                if imported_metadata.id not in visited:
                    stack.append((imported, False))

        return post_order

    def _register_modules(self, modules: list[tuple[ModuleType, ModuleMetadata]]) -> Module:
        for type_, metadata in modules:
            if metadata.id in self._modules:
                continue
            self._register_module(type_, metadata)
        _, root_metadata = self._compiler.extract_metadata(self._root_module_type)
        return self._modules[root_metadata.id]

    def _register_module(self, type_: ModuleType, metadata: ModuleMetadata) -> Module:
        module = Module(type_, metadata)
        self._modules[module.id] = module
        self._providers.extend(module.providers)
        self._update_adjacency(module)
        return module

    def _handle_extensions(self, metadata: ModuleMetadata) -> None:
        for extension in metadata.extensions:
            if isinstance(extension, OnModuleConfigure):
                extension.on_module_configure(metadata)
            self._extensions.append(extension)

    def _update_adjacency(self, module: Module) -> None:
        if module.id not in self._adjacency:
            self._adjacency[module.id][module.id] = str(module)

        for imported_module_type in module.imports:
            _, imported_metadata = self._compiler.extract_metadata(imported_module_type)
            if imported_metadata.id not in self._adjacency[module.id]:
                self._adjacency[module.id][imported_metadata.id] = str(imported_metadata)

    def _build_registry(self, root_module: Module) -> ModuleRegistry:
        return ModuleRegistry(
            compiler=self._compiler,
            modules=self._modules,
            providers=self._providers,
            extensions=self._extensions,
            root_module=root_module,
            adjacency=self._adjacency,
        )



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from collections.abc import Sequence
    from uuid import UUID

    from waku.di import BaseProvider
    from waku.extensions import ModuleExtension
    from waku.modules._metadata import DynamicModule, ModuleMetadata, ModuleType


__all__ = ['Module']


class Module:
    def __init__(self, module_type: ModuleType, metadata: ModuleMetadata) -> None:
        self.id: Final[UUID] = metadata.id
        self.target: Final[ModuleType] = module_type
        self.distance: Final[int] = sys.maxsize if metadata.is_global else 0

        self.providers: Final[Sequence[BaseProvider]] = metadata.providers
        self.imports: Final[Sequence[ModuleType | DynamicModule]] = metadata.imports
        self.exports: Final[Sequence[object | ModuleType]] = metadata.exports
        self.extensions: Final[Sequence[ModuleExtension]] = metadata.extensions
        self.is_global: Final[bool] = metadata.is_global

    @property
    def name(self) -> str:
        return self.target.__name__

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f'Module[{self.name}]'

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Module):  # pragma: no cover
            return False
        return self.id == other.id

    def __lt__(self, other: Module) -> bool:
        return self.distance < other.distance



from __future__ import annotations

from dataclasses import dataclass

from waku import WakuFactory, module
from waku.di import (
    contextual,
    object_,
    scoped,
    singleton,
    transient,
)


@dataclass
class Service:
    value: int = 1


@dataclass
class DependentService:
    service: Service


@dataclass
class RequestContext:
    user_id: int


@dataclass
class UserService:
    user_id: int


def create_user_service(ctx: RequestContext) -> UserService:
    return UserService(ctx.user_id)


async def test_provider_scopes() -> None:
    @module(
        providers=[
            scoped(DependentService),
            transient(lambda: Service(value=2), provided_type=Service),
        ]
    )
    class AppModule:
        pass

    app = WakuFactory(AppModule).create()

    async with app, app.container() as container:
        # Test transient - new instance each time
        service1 = await container.get(Service)
        service2 = await container.get(Service)
        assert service1 is not service2
        assert service1.value == 2
        assert service2.value == 2

        # Test scoped - same instance per request
        dep1 = await container.get(DependentService)
        dep2 = await container.get(DependentService)
        assert dep1 is dep2


async def test_object_provider() -> None:
    service = Service(value=3)

    @module(providers=[object_(service, provided_type=Service)])
    class AppModule:
        pass

    app = WakuFactory(AppModule).create()

    async with app, app.container() as container:
        result = await container.get(Service)
        assert result is service
        assert result.value == 3


async def test_contextual_provider() -> None:
    service_instance = Service(value=4)

    @module(providers=[contextual(Service)])
    class AppModule:
        pass

    app = WakuFactory(AppModule).create()

    async with app:
        context = {Service: service_instance}
        async with app.container(context=context) as container:
            result = await container.get(Service)
            assert result is service_instance
            assert result.value == 4


async def test_request_context_provider() -> None:
    @module(
        providers=[
            contextual(RequestContext),
            scoped(create_user_service, provided_type=UserService),
        ]
    )
    class AppModule:
        pass

    app = WakuFactory(AppModule).create()

    async with app:
        # First request
        context1 = {RequestContext: RequestContext(user_id=1)}
        async with app.container(context=context1) as container1:
            user1 = await container1.get(UserService)
            assert user1.user_id == 1

        # Second request with different context
        context2 = {RequestContext: RequestContext(user_id=2)}
        async with app.container(context=context2) as container2:
            user2 = await container2.get(UserService)
            assert user2.user_id == 2


async def test_provider_with_type() -> None:
    class BaseService:
        pass

    class ConcreteService(BaseService):
        pass

    @module(providers=[scoped(ConcreteService, provided_type=BaseService)])
    class AppModule:
        pass

    app = WakuFactory(AppModule).create()

    async with app, app.container() as container:
        result = await container.get(BaseService)
        assert isinstance(result, ConcreteService)


async def test_injected_dependency() -> None:
    @module(
        providers=[
            singleton(lambda: 1, provided_type=int),
            scoped(Service),
            scoped(DependentService),
        ]
    )
    class AppModule:
        pass

    app = WakuFactory(AppModule).create()

    async with app, app.container() as container:
        service = await container.get(Service)
        dep = await container.get(DependentService)
        assert dep.service is service



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING

from waku import WakuApplication, WakuFactory
from waku.di import scoped, singleton
from waku.extensions import AfterApplicationInit, OnApplicationInit, OnModuleConfigure, OnModuleInit
from waku.modules import Module, ModuleMetadata, module

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from pytest_mock import MockerFixture


class MockLifespanManager:
    def __init__(self) -> None:
        self.entered = False
        self.exited = False

    @asynccontextmanager
    async def __call__(self, _: WakuApplication) -> AsyncIterator[None]:
        self.entered = True
        yield
        self.exited = True


async def test_application_lifespan() -> None:
    """Test application startup and shutdown lifecycle."""
    manager_1 = MockLifespanManager()
    manager_2 = MockLifespanManager()

    @module()
    class AppModule:
        pass

    application = WakuFactory(
        AppModule,
        lifespan=[manager_1, manager_2],
    ).create()

    async with application:
        assert manager_1.entered
        assert manager_2.entered
        assert not manager_1.exited
        assert not manager_2.exited

    assert manager_1.exited
    assert manager_2.exited  # type: ignore[unreachable]


async def test_application_module_registration() -> None:
    """Test that modules are properly registered with the application."""

    class ServiceA:
        pass

    class ServiceB:
        pass

    @module(providers=[scoped(ServiceA)], exports=[ServiceA])
    class ModuleA:
        pass

    @module(providers=[singleton(ServiceB)], imports=[ModuleA])
    class ModuleB:
        pass

    @module(imports=[ModuleA, ModuleB])
    class AppModule:
        pass

    application = WakuFactory(AppModule).create()

    async with application, application.container() as request_container:
        await request_container.get(ServiceA)
        await application.container.get(ServiceB)


async def test_on_module_init_extension_called(mocker: MockerFixture) -> None:
    """Test that Module extension is called for the module."""
    on_module_configure_mock = mocker.MagicMock()
    on_module_init_mock = mocker.AsyncMock()

    class ModuleOnConfigureExt(OnModuleConfigure):
        def on_module_configure(self, metadata: ModuleMetadata) -> None:  # noqa: PLR6301
            on_module_configure_mock(metadata)

    class ModuleOnInitExt(OnModuleInit):
        async def on_module_init(self, module: Module) -> None:  # noqa: PLR6301
            await on_module_init_mock(module)

    @module(
        extensions=[
            ModuleOnConfigureExt(),
            ModuleOnInitExt(),
        ],
    )
    class AppModule:
        pass

    application = WakuFactory(AppModule).create()
    await application.initialize()

    assert on_module_configure_mock.call_count == 1
    assert isinstance(on_module_configure_mock.call_args[0][0], ModuleMetadata)
    assert on_module_init_mock.call_count == 1
    assert isinstance(on_module_init_mock.call_args[0][0], Module)


async def test_application_init_extensions_called(mocker: MockerFixture) -> None:
    on_app_init_mock = mocker.AsyncMock()
    after_app_init_mock = mocker.AsyncMock()

    class AppOnInitExt(OnApplicationInit):
        async def on_app_init(self, app: WakuApplication) -> None:  # noqa: PLR6301
            await on_app_init_mock(app)

    class AppAfterInitExt(AfterApplicationInit):
        async def after_app_init(self, app: WakuApplication) -> None:  # noqa: PLR6301
            await after_app_init_mock(app)

    @module()
    class AppModule:
        pass

    application = WakuFactory(
        AppModule,
        extensions=[
            AppOnInitExt(),
            AppAfterInitExt(),
        ],
    ).create()

    # Should be called once for the application initialization
    await application.initialize()
    await application.initialize()

    assert on_app_init_mock.call_count == 1
    assert isinstance(on_app_init_mock.call_args[0][0], WakuApplication)
    assert after_app_init_mock.call_count == 1



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any

from dishka.entities.factory_type import FactoryType
from typing_extensions import override

from waku.di import AsyncContainer, Scope
from waku.ext.validation._abc import ValidationRule
from waku.ext.validation._errors import ValidationError

if TYPE_CHECKING:
    from waku.ext.validation._extension import ValidationContext
    from waku.modules import Module, ModuleRegistry

__all__ = ['DependenciesAccessible']


class DependenciesAccessible(ValidationRule):
    """Check if all dependencies of providers are accessible.

    This validation rule ensures that all dependencies required by providers
    are either:
    1. Available globally
    2. Provided by the current module
    3. Provided by any of the imported modules
    """

    @override
    def validate(self, context: ValidationContext) -> list[ValidationError]:
        registry: ModuleRegistry = context.app.registry
        modules: list[Module] = list(registry.traverse())  # Validate all registered modules

        # Cache global providers
        global_providers: set[type[object]] = {AsyncContainer}
        global_providers |= {
            provided_type
            for module in modules
            for provided_type in _module_provided_types(module)
            if registry.is_global_module(module)
        }
        global_context_providers = {
            dep.type_hint
            for dep, factory in context.app.container.registry.factories.items()
            if (factory.scope is Scope.APP and factory.type is FactoryType.CONTEXT)
        }
        global_providers |= global_context_providers

        errors: list[ValidationError] = []
        for module in modules:
            imported_modules = [m for m in registry.traverse(module) if m != module]
            for provider in module.providers:
                for factory in provider.factories:
                    inaccessible_deps: set[type[object]] = set()
                    for dependency in factory.dependencies:
                        dep_type = dependency.type_hint
                        if not _can_access_dependency(dep_type, module, global_providers, imported_modules):
                            inaccessible_deps.add(dep_type)

                    for dep_type in inaccessible_deps:
                        err_msg = (
                            f'"{factory.source!r}" from "{module!r}" depends on "{dep_type!r}" but it\'s not accessible to it\n'
                            f'To resolve this issue:\n'
                            f'   1. Export "{dep_type!r}" from some module\n'
                            f'   2. Add module which exports "{dep_type!r}" to "{module!r}" imports'
                        )
                        errors.append(ValidationError(err_msg))

        return errors


def _can_access_dependency(
    dep_type: Any,
    module: Module,
    global_providers: set[type[object]],
    imported_modules: list[Module],
) -> bool:
    # 1. Dep available globally
    if dep_type in global_providers:
        return True
    # 2. Dep provided by current module (regardless of export)
    if dep_type in _module_provided_types(module):
        return True
    # 3. Dep extracted from container context
    if dep_type in _module_provided_context_vars(module):
        return True
    # 4. Dep provided by any of imported modules (must be exported)
    return any(_is_type_exported_by_module(dep_type, imported_module) for imported_module in imported_modules)


@functools.cache
def _is_type_exported_by_module(type_: type[object], module: Module) -> bool:
    """Check if a type is exported by a module."""
    return type_ in _module_provided_types(module) and type_ in module.exports


@functools.cache
def _module_provided_types(module: Module) -> set[type[object]]:
    """Get all types provided by a module's providers.

    Args:
        module: The module to extract provided types from

    Returns:
        A set of all type hints that the module's providers can provide
    """
    # fmt: off
    return {
        factory.provides.type_hint
        for provider in module.providers
        for factory in provider.factories
    }
    # fmt: on


@functools.cache
def _module_provided_context_vars(module: Module) -> set[type[object]]:
    """Get all types provided by a provider's context variables."""
    # fmt: off
    return {
        context_var.provides.type_hint
        for provider in module.providers
        for context_var in provider.context_vars
    }
    # fmt: on



from dataclasses import dataclass

from waku import DynamicModule, WakuFactory, module
from waku.di import scoped
from waku.extensions import OnModuleConfigure
from waku.modules import ModuleMetadata


@dataclass
class Dep:
    pass


def test_on_module_configure_extension_changes_module_metadata_only_once() -> None:
    class AddDepOnConfigure(OnModuleConfigure):
        def on_module_configure(self, metadata: ModuleMetadata) -> None:  # noqa: PLR6301
            metadata.providers.append(scoped(Dep))

    @module(
        providers=[],
        extensions=[AddDepOnConfigure()],
    )
    class SomeModule:
        pass

    @module(imports=[SomeModule])
    class AppModule:
        pass

    WakuFactory(AppModule).create()
    application = WakuFactory(AppModule).create()

    assert len(application.registry.get_by_type(SomeModule).providers) == 1


def test_on_module_configure_extension_with_dynamic_module() -> None:
    class AddDepOnConfigure(OnModuleConfigure):
        def on_module_configure(self, metadata: ModuleMetadata) -> None:  # noqa: PLR6301
            metadata.providers.append(scoped(Dep))

    @module()
    class SomeModule:
        @classmethod
        def register(cls) -> DynamicModule:
            return DynamicModule(
                parent_module=cls,
                extensions=[AddDepOnConfigure()],
            )

    dynamic_module = SomeModule.register()

    @module(imports=[dynamic_module])
    class AppModule:
        pass

    application = WakuFactory(AppModule).create()

    assert len(application.registry.get_by_type(dynamic_module).providers) == 1


def test_module_imported_twice_only_once() -> None:
    class AddDepOnConfigure(OnModuleConfigure):
        def on_module_configure(self, metadata: ModuleMetadata) -> None:  # noqa: PLR6301
            metadata.providers.append(scoped(Dep))

    @module(
        providers=[],
        extensions=[AddDepOnConfigure()],
    )
    class SomeModule:
        pass

    @module(imports=[SomeModule])
    class ModuleA:
        pass

    @module(imports=[SomeModule])
    class ModuleB:
        pass

    @module(imports=[ModuleA, ModuleB])
    class AppModule:
        pass

    application = WakuFactory(AppModule).create()
    # Ensure that configuration for SomeModule is applied only once despite multiple import paths.
    assert len(application.registry.get_by_type(SomeModule).providers) == 1



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections import OrderedDict
    from collections.abc import Iterator
    from uuid import UUID

    from waku.di import BaseProvider
    from waku.extensions import ModuleExtension
    from waku.modules._metadata import DynamicModule, ModuleCompiler, ModuleType
    from waku.modules._module import Module

__all__ = ['ModuleRegistry']


class ModuleRegistry:
    """Immutable registry and graph for module queries, traversal, and lookups."""

    def __init__(
        self,
        *,
        compiler: ModuleCompiler,
        root_module: Module,
        modules: dict[UUID, Module],
        providers: list[BaseProvider],
        extensions: list[ModuleExtension],
        adjacency: dict[UUID, OrderedDict[UUID, str]],
    ) -> None:
        self._compiler = compiler
        self._root_module = root_module
        self._modules = modules
        self._providers = tuple(providers)
        self._extensions = tuple(extensions)
        self._adjacency = adjacency

    @property
    def root_module(self) -> Module:
        return self._root_module

    @property
    def modules(self) -> tuple[Module, ...]:
        return tuple(self._modules.values())

    @property
    def providers(self) -> tuple[BaseProvider, ...]:
        return self._providers

    @property
    def compiler(self) -> ModuleCompiler:
        return self._compiler

    def get(self, module_type: ModuleType | DynamicModule) -> Module:
        module_id = self._compiler.extract_metadata(module_type)[1].id
        return self.get_by_id(module_id)

    def get_by_id(self, module_id: UUID) -> Module:
        module = self._modules.get(module_id)
        if module is None:
            msg = f'Module with ID {module_id} is not registered in the graph.'
            raise KeyError(msg)
        return module

    def get_by_type(self, module_type: ModuleType | DynamicModule) -> Module:
        _, metadata = self._compiler.extract_metadata(module_type)
        return self._modules[metadata.id]

    def traverse(self, from_: Module | None = None) -> Iterator[Module]:
        """Traverse the module graph in depth-first post-order (children before parent) using a stack.

        Args:
            from_: Start module (default: root)

        Yields:
            Module: Each traversed module (post-order)
        """
        start_module = from_ or self._root_module
        visited: set[UUID] = set()
        stack: list[tuple[Module, bool]] = [(start_module, False)]

        while stack:
            module, processed = stack.pop()

            if processed:
                yield module
                continue

            if module.id in visited:
                continue

            visited.add(module.id)

            # Push the current module to process after its children
            stack.append((module, True))

            # Push children to be processed first (in reverse order to maintain original order in DFS)
            neighbor_ids = self._adjacency[module.id]
            for neighbor_id in reversed(neighbor_ids.keys()):
                if neighbor_id == module.id:
                    continue
                neighbor = self.get_by_id(neighbor_id)
                stack.append((neighbor, False))

    def is_global_module(self, module: Module) -> bool:
        return module.is_global or module == self._root_module
